/*
 * Brute force attack using a hash table based on source IP.
 * We will not check the packets which are too close in time.
 * https://en.wikipedia.org/wiki/Jenkins_hash_function
 */

#define _GNU_SOURCE

#include <sys/time.h>
#include <sys/random.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "util.h"
#include "brute.h"

static unsigned char	secret[8];
static unsigned int	bad_entries;
static time_t		*brute_table;

static char		error[256];

char *brute_error(void)
{
	return error;
}

/*
 * Hash function
 */
static unsigned int brute_hash(const unsigned char *src,
	const unsigned char src_len)
{
	unsigned char i = 0;
	unsigned int hash = 0;

	while (i != src_len) {
		hash += src[i++];
		hash += hash << 10;
		hash ^= hash >> 6;
	}

	// mix secret
	i = 0;
	while (i < sizeof(secret)) {
		hash += secret[i++];
		hash += hash << 10;
		hash ^= hash >> 6;
	}

	hash += hash << 3;
	hash ^= hash >> 11;
	hash += hash << 15;

	return hash;
}

/*
 * Init brute force system
 */
int brute_init(const unsigned int entries)
{
	int r;

	if (entries == 0) {
		snprintf(error, sizeof(error),
			"bad entries size cannot be zero");
		return -1;
	}

	bad_entries = entries;

	brute_table = malloc(bad_entries * sizeof(time_t));
	if (!brute_table) {
		snprintf(error, sizeof(error),
			"cannot alloc memory for bad entries");
		return -1;
	}

	memset(brute_table, 0, bad_entries * sizeof(time_t));

	r = my_getrandom(secret, sizeof(secret), 0);
	if (r == -1) {
		snprintf(error, sizeof(error),
			"cannot get random bytes to populate the secret");
		free(brute_table);
		return -1;
	}

	return 0;
}

/*
 * Look up an IP, and if not present, store it
 * Returns 0 if IP was not present or it is ok to check it again.
 * Returns -1 if we just need to drop the packet.
 */
int brute_check(const unsigned char *src, const unsigned char src_len,
	const time_t now)
{
	unsigned int h;

	h = brute_hash(src, src_len) % bad_entries;

	if (brute_table[h] + 10 < now) {
		brute_table[h] = now;
		return 0;
	}

	return -1;
}
